# passerelle - uniform access to multiple data sources and services
# Copyright (C) 2019  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import copy
import datetime
import logging
import os
import re
import xml.etree.ElementTree as ET


from django.db import models
from django.http import HttpResponse
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

import xmlschema

from passerelle.base.models import BaseResource
from passerelle.utils.api import endpoint
from passerelle.utils.jsonresponse import APIError
from passerelle.utils.xml import JSONSchemaFromXMLSchema

logger = logging.getLogger('passerelle.contrib.toulouse_axel')

BASE_XSD_PATH = os.path.join(os.path.dirname(__file__), 'xsd')


class AxelSchema(JSONSchemaFromXMLSchema):
    type_map = {
        '{urn:AllAxelTypes}DATEREQUIREDType': 'date',
        '{urn:AllAxelTypes}DATEType': 'date',
    }

    @classmethod
    def schema_date(cls):
        return {
            'type': 'string',
            'pattern': '[0-9]{4}-[0-9]{2}-[0-9]{2}',
        }

    def encode_date(self, obj):
        return datetime.datetime.strptime(obj, '%Y-%m-%d').strftime('%d/%m/%Y')

    def decode_date(self, data):
        if not data.text:
            return data
        value = datetime.datetime.strptime(data.text, '%d/%m/%Y').strftime('%Y-%m-%d')
        return xmlschema.ElementData(tag=data.tag, text=value, content=data.content, attributes=data.attributes)


class AxelError(Exception):
    pass


def xml_schema_converter(name, root_element):
    return AxelSchema(os.path.join(BASE_XSD_PATH, name), root_element)


class Operation(object):
    def __init__(self, operation, prefix='Dui/'):
        self.operation = operation
        self.request_converter = xml_schema_converter('%sQ_%s.xsd' % (prefix, operation), 'PORTAIL')
        self.response_converter = xml_schema_converter('%sR_%s.xsd' % (prefix, operation), 'PORTAILSERVICE')
        self.name = re.sub(
            '(.?)([A-Z])',
            lambda s: s.group(1) + ('-' if s.group(1) else '') + s.group(2).lower(),
            operation)
        self.snake_name = self.name.replace('-', '_')

    @property
    def request_schema(self):
        schema = self.request_converter.json_schema
        schema['flatten'] = True
        schema['merge_extra'] = True
        return schema

    def __call__(self, resource, request_data):
        client = resource.soap_client()
        serialized_request = ET.tostring(
            self.request_converter.encode(request_data))
        result = client.service.getData(
            self.operation,
            force_text(serialized_request),
            '')  # FIXME: What is the user parameter for ?

        xml_result = ET.fromstring(result)
        if xml_result.find('RESULTAT/STATUS').text != 'OK':
            msg = xml_result.find('RESULTAT/COMMENTAIRES').text
            raise AxelError(msg)

        try:
            return self.response_converter.decode(ET.fromstring(result))
        except xmlschema.XMLSchemaValidationError as e:
            raise AxelError('invalid response', e)


ref_verif_dui = Operation('RefVerifDui')
ref_famille_dui = Operation('RefFamilleDui')
ref_facture_a_payer = Operation('RefFactureAPayer')
ref_facture_pdf = Operation('RefFacturePDF', prefix='')


class ToulouseAxel(BaseResource):

    wsdl_url = models.CharField(
        max_length=128,
        blank=False,
        verbose_name=_('WSDL URL'),
        help_text=_('Toulouse Axel WSDL URL'))

    category = _('Business Process Connectors')

    class Meta:
        verbose_name = _('Toulouse Axel')

    def check_status(self):
        response = self.requests.get(self.wsdl_url)
        response.raise_for_status()

    LINK_SCHEMA = copy.deepcopy(ref_verif_dui.request_schema['properties']['PORTAIL']['properties']['DUI'])
    LINK_SCHEMA['properties'].pop('IDPERSONNE')
    LINK_SCHEMA['required'].remove('IDPERSONNE')

    @endpoint(
        description=_('Create link between user and Toulouse Axel'),
        perm='can_access',
        parameters={
            'NameID': {'description': _('Publik ID')},
        },
        post={
            'request_body': {
                'schema': {
                    'application/json': LINK_SCHEMA,
                }
            }
        })
    def link(self, request, NameID, post_data):
        if not NameID:
            raise APIError('NameID is empty', err='bad-request', http_status=400)

        post_data['IDPERSONNE'] = ''
        try:
            result = ref_verif_dui(self, {'PORTAIL': {'DUI': post_data}})
        except AxelError as e:
            raise APIError('Axel error: %s' % e, err='error')

        dui_data = result['DATA']['PORTAIL']['DUI']
        code = dui_data['CODE']
        if code not in [2, 3]:
            # 2: RL1; 3: RL2
            raise APIError('Person not found', err='not-found')
        link, created = self.link_set.get_or_create(
            name_id=NameID,
            defaults={
                'dui': dui_data['IDDUI'],
                'person_id': dui_data['IDPERSONNE']})
        if not created and (link.dui != dui_data['IDDUI'] or link.person_id != dui_data['IDPERSONNE']):
            raise APIError('Data conflict', err='conflict')
        return {'link': link.pk, 'created': created, 'dui': link.dui}

    def get_link(self, name_id, error_status=200):
        try:
            return self.link_set.get(name_id=name_id)
        except Link.DoesNotExist:
            raise APIError('Person not found', err='not-found', http_status=error_status)

    @endpoint(
        description=_('Delete link between user and Toulouse Axel'),
        methods=['post'],
        perm='can_access',
        parameters={
            'NameID': {'description': _('Publik ID')},
        })
    def unlink(self, request, NameID):
        link = self.get_link(NameID)
        link_id = link.pk
        link.delete()
        return {'link': link_id, 'deleted': True, 'dui': link.dui}

    def get_family_data(self, name_id):
        link = self.get_link(name_id)
        try:
            result = ref_famille_dui(self, {'PORTAIL': {'DUI': {'IDDUI': link.dui}}})
        except AxelError as e:
            raise APIError('Axel error: %s' % e, err='error')
        return result['DATA']['PORTAIL']['DUI']

    @endpoint(
        description=_("Get information about user's family"),
        perm='can_access',
        parameters={
            'NameID': {'description': _('Publik ID')},
        })
    def family_info(self, request, NameID):
        family_data = self.get_family_data(NameID)
        return {'data': family_data}

    @endpoint(
        description=_("Get information about a child"),
        perm='can_access',
        parameters={
            'NameID': {'description': _('Publik ID')},
            'idpersonne': {'description': _('Child ID')},
        })
    def child_info(self, request, idpersonne, NameID):
        family_data = self.get_family_data(NameID)

        for child in family_data['ENFANT']:
            if child['IDPERSONNE'] == idpersonne:
                return {'data': child}

        raise APIError('Child not found', err='not-found')

    def get_invoices(self, name_id, error_status=200):
        link = self.get_link(name_id, error_status=error_status)
        try:
            result = ref_facture_a_payer(self, {'PORTAIL': {'DUI': {'IDDUI': link.dui}}})
        except AxelError as e:
            raise APIError('Axel error: %s' % e, err='error', http_status=error_status)
        data = result['DATA']['PORTAIL']['DUI']
        result = []
        for facture in data.get('FACTURES', []):
            result.append({
                'id': facture['IDFACTURE'],
                'label': facture['LIBELLE'],
                'amount': facture['RESTEAPAYER'],
                'total_amount': facture['MONTANTTOTAL'],
                'online_payment': True,
                'created': facture['DATEEMISSION'],
                'pay_limit_date': facture['DATEECHEANCE'],
                'has_pdf': True if facture['EXISTEPDF'] == '1' else False,
                'paid': False,
                'vendor': {'toulouse-axel': facture},
            })
        return result

    def get_invoice(self, name_id, invoice_id, error_status=200):
        invoices_data = self.get_invoices(name_id, error_status=error_status)
        for invoice in invoices_data:
            if str(invoice['id']) == invoice_id:
                return invoice

    @endpoint(
        description=_("Get invoices to pay"),
        perm='can_access',
        parameters={
            'NameID': {'description': _('Publik ID')},
        })
    def invoices(self, request, NameID):
        invoices_data = self.get_invoices(NameID)
        return {'data': invoices_data}

    @endpoint(
        perm='can_access',
        pattern=r'^(?P<invoice_id>\w+)/?$',
        description=_('Get invoice details'),
        parameters={
            'NameID': {'description': _('Publik ID')},
            'invoice_id': {'description': _('Invoice identifier')}
        })
    def invoice(self, request, invoice_id, NameID):
        invoice = self.get_invoice(NameID, invoice_id)
        if invoice is None:
            raise APIError('Invoice not found', err='not-found')

        return {'data': invoice}

    @endpoint(
        name='invoice',
        perm='can_access',
        pattern=r'^(?P<invoice_id>\w+)/pdf/?$',
        description=_('Get invoice as a PDF file'),
        parameters={
            'NameID': {'description': _('Publik ID')},
            'invoice_id': {'description': _('Invoice identifier')}
        })
    def invoice_pdf(self, request, invoice_id, NameID):
        # check that invoice is related to current user
        invoice = self.get_invoice(NameID, invoice_id, error_status=404)
        if invoice is None:
            raise APIError('Invoice not found', err='not-found', http_status=404)
        # check that PDF is available
        if not invoice['has_pdf']:
            raise APIError('PDF not available', err='not-available', http_status=404)

        try:
            result = ref_facture_pdf(self, {'PORTAIL': {'FACTUREPDF': {'IDFACTURE': int(invoice_id)}}})
        except AxelError as e:
            raise APIError('Axel error: %s' % e, err='error', http_status=404)

        b64content = base64.b64decode(result['DATA']['PORTAIL']['PDF']['@FILE'])
        if not b64content:
            raise APIError('PDF error', err='error', http_status=404)
        response = HttpResponse(content_type='application/pdf')
        response['Content-Disposition'] = 'attachment; filename="%s.pdf"' % invoice_id
        response.write(b64content)
        return response


class Link(models.Model):
    resource = models.ForeignKey(ToulouseAxel, on_delete=models.CASCADE)
    name_id = models.CharField(blank=False, max_length=256)
    dui = models.CharField(blank=False, max_length=128)
    person_id = models.CharField(blank=False, max_length=128)

    class Meta:
        unique_together = ('resource', 'name_id')
